from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"author": {"first_name": "Patrick", "last_name": "Rothfuss"}, "title": "THE NAME OF THE WIND", "rating": 7}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                        "title": "THE NAME OF THE WIND",
                                        "rating": 7,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n\xd8\x05\x01\xd1\xed\x8ao\xd7(\x17x2\xee;\xdc<\x1ap8\x13\x88e\xd6\x99Cz\x9d\x88\xfbq\x82\x84\xd4\xeb\xc6Qp\xe7\xfc\xa0\xd0\xa0\xa7N\xfeC\xf6\n\xf5\x88\xdf\xb1\xfd\x82a\x93M2v.@:Q\xe9\xde\x11\xd910\xadK\x10f\x90\xee\x10o\xffz\xdc\xb0R\x82>\xb1\x1d-\x8d\xe7\x9e;\xd9~\xbf\xfc\x89\x84)\x1a\xaetf\xfe\x8b\x01\xc0\x83\xb9J\x97`\x86\x03\xab\xd2w\x81c \xe43\xb8l\xf4?@!\xa1\x1e3\xef\x86\xf9f\x1d\xd9Tl\xe9\x1b\x17\x9b~\xc8\xb5\xb1\x9e\x01\x8e\xdc\x0c\xec\x97[4\xf2)\x02^\xa7%\x8f3\x97)2+\xa6\xd0\x0b@*\x82\xe91\xab\xe6M\x0b\xcb\xe4\x85\x9b\xefy\xf4\x98\x1bV\x9b!\xc4a\x84\x18\xfe\xc3\xda\xeb\x12%\xffL\x0c\xb2\xb7\x15 \xee\x83\x7f\x01C\xf4\x11Y6\xf0\xadS\xe5T`\xcf<g\xdd\xbf\x99\xfd\x8d\x90\xeb\xe2\r\x1cV6\x8e\xc0k\r@\xce\x9b\xfd\x06d\xef\xe4\xef\xba\x94\xc9\xff\x86\x96\xc3\xdc'\xd9\xfe2\x96(\x9bS\xf7D\xc3%r\xba\xd7q\xd2p\xcf\xc0E\x83\x8c\xf0\x05\x03\xac\xe7M\x19\xb6R\x91\x15\xa0\x8cT\x81w1\xe9\xe1Z\x81\x1e\x02\xa5\xf5\x18\x8b\x115\xda\x1alH\xfc\xd2Z\xaa2\x9b\xdb2\x0e\x1b\x19\x10\xd7\xdb\xb9\xc7\xed\x95\xc1\x87\x1e\xac\xabn\xe3\xb3\xeb\xec\xe3c\xecOd\xa3\xbf\x13\xf4=\t\x8d\xdc|\xa4\xb9;\xb3\xf1\x06u.u \xbb\"\xb6I\x9f\xe4cQ\x16R\x9d\xde\x97\xde\xce\xe8\xb0\xf5gWT!\xee\xcc\xe3N\r\x8d\xb6\x08\xc4\x1f_l\xbbcv\x85\x16\xde\x0c[T\x05`\xc4AN7\xa7?\xc7\x16 \xcc\x0c\xc2\xf1\xdeZ\x8e~\xc7\x9f\x84\x08\xda\x00\x08\x94\xbf\x91`)S*=N\xdcAg\x1b\xa5N\xbc\xb5Q\n\x0cn\t{\x8a@\xb8B\x13\x9ak,\x19\xef\xe4`F%\xa2C\xc3\x1e\xc0&\x01\x92\xe5\xeam\x8e\xd8\xee\x06\xff@\xf8\xb8\x1d\x17\xef\x92@\x9f\x08\xd9\xef3\xd2ea\xf4,fD\xfb\xe8\x1e@\x1b\xad\x9d\xab\x1e\xb1\x0bB\x8fy\x0cv\xc49/\xff\x13\x1c\xdeT\xe1\\\xa0\xdaf\x8dU\xe2P{\xe7\x84\x92\xfc~\x12$\xf4\x10\x93\xb4\x0e\xae_\x96\xc8\xf8d\xd5\xa7\xff\xafy\x04\x00\xc7\x8e\xf3\x1c\x9e\x90PW\x90\xd1\x84lM@\xc1\xa6&3\x91\xc8\x99,\xff\x0bWH%\xed\xbc\x04Y\xf9\x9a\x1d7\xf9`\xb6\xcf\xff\x94\x81N\x99\xfb\xa39\xd4\x84@\xecszc\xe2\x1a\xc2J\xac\xfa\xe5\xf3#\nN&D\x01\x85r\xa1\x17\xf5>1\xe1,U\xb5c\xbf\x9a.F\xae\x01\x03\x00\xa5\xc7$V\xae\x1b\x12Y[\xb7n\xd1\x94\xf9\xa1\x03N\xda\xa7\xbe\x8dL{\x82\xb0\x184d\xe7\xf6\xfbu\xa0\xfd\xc2*\x01n\xf5\x8fzh?\xa5~d+W\xbc\xe6\xb1!\xb4\xbbm\xc7U\x9e\xb9\xb3H\xc2\x14LoE\xc1\x06\n\xd0\xffu\xfb\xe3",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"rating": 7, "author": {"first_name": "Patrick", "last_name": "Rothfuss"}, "title": "THE NAME OF THE WIND"}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "rating": 7,
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                        "title": "THE NAME OF THE WIND",
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n\x94\x05\x01\xd1\xed\x8aoKY}\xbf<I\xb4\x0f\x7f({\xd4~2{\x0e\xf9\x0c\x1bB<\x13C\x9b\x98\x040\x19\xba-\xb4\x04\x86\x98\x1eX\x0c\xc6\xed\xb5\xe3JG\x17\r\x89\xc9\xecL\xa6\xb3H\xae\x96*\xe6\xceY\xad\xa8K\x1e\x82\x08E\x93hS\x044tf\xd3m \x10\x9f\x02#\x88\xfb\x8b\xf6]]\xc7\xfd\xa3S\x04\x12u_\x85 \xe5V\x99\xe2\x80\x17#\x85\xb2\xe0\x92;\xf3\x81\x04\x10\x86V;+!^\x0ep\xfd'{L\xc3w\r_\x16[\x18\xeb\xdf\x05/\xce\x1bpf\xb7r\r\xba\xdb\xb6D\x9blm\xef\x96\xa79\x00h\x01_\x13g\xb9\x98\x0bB0\x1dg\xcb:\xfb\xe9M\x9d|Z\"\xf3/\xa0o\x82R\xb2\xc07\xac\xdem+>\xd6W\xe8u\xb9.A>\xed7n\x80\xc6\xcbk\xde.x\xb9\r\x86\xff \xf8a\xb5D\x06%\xa6Zt\xa1t`\xb7=\xd4\xfd\x80\x8b.3\x1ba\xc3\x91\x16\xc9{\x87\xddv\xc6*B\xf8\x85MO\xb3\xe3\x88\x9e\xa6h6q\xf19\xa2\x90`I\xd1ya\x99\xf1\xb3(B\xaf\x9e\x18\xbe\x1a4@\xb6\x9d\x0e\xe1}\xca\xd6\xb0\\\x06Vr[\x8a\t\xf3\x86\xcb\x7f\x0eJb-Sh\r\x14\n;\x9b\xa1\t\xfe+\x17}\xb5\x90\xe5\x06\xdf@fb\x9d\xbboF\xe1\xbd\xb1\xa7\x9a\xed\xc8\x8d\xden\xae\xba\xab\xfer\x1f0(\x9c\xe1f&\xeb\xf9)\xb4>\tt\x02\xb3<{GE9?\x0e}\x9c\xcf^g\xc4E\xc7\xa1\xb1U\x1e\xd3\x1d:\xe1\xaec}\xdb\xd8\xd4\xac\xe8\xbfWN\xb4yh%\x9c\x97.\xc3\x1e^\xab\x82\xd2\x18\xb45\x9e.f\x83}\xc1=rI>\xdf\xa2\xa5.\xab\x0f\xd8O\xd6%\x90B\xea\x8b[\xd7+\x8b\xa2B\xfe\x0beY\xa7o\xbew\xd1\xe1\xc1v\xc4AX\xef\xf8\xd7\xecWx\xf3:\xb9^g[~\x90\xf2\xf3\xe7p\xe5,\x01\xf1\x85z\xf1D\x9c@\xd7\x99#&\xa8\xfb\xba\xd8\xe6y\xd9'4\x14\xd05b\xd5:\xa5QB\xf1\x98q\xe6X\x86\xff\x10;\xddw\xc4I\xab\xdeE\xf1\x9dn\xda\x84bk\x91\xc1\xe8\xeaT\x91}\xecD\xb7u\xf2(R!?\xba\x96\xac\x8a\xfe5\xc1\x13t\xc0\xb9\x82\n\xba\t%\x06\x9d\xe9\x02\xd4<\xb3\xe3b\xf2B\xc9\x1aZ'j\xa1J\x8e\xd9\xd3\xce\xec\xe5Wu=\x83v\x92\x1c\xbe\x94\"g\xd8ca\x01k\xf8hn\x9c\xd4v\x9f\x9d0.7\xd9\xd7\x8b\xe0\x7fv\xe4\xe7Yq \xae\\\x153\xcf\x8d_sx0\xa72H\x0e\x11p?\xe3\xffR2~\x10\xe7\xc7>.M\xc9\x92t6PlQ\x8c\xb2",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"author": {"first_name": "Patrick", "last_name": "Rothfuss"}, "title": "THE NAME OF THE WIND", "rating": 7}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "author": {
                                            "first_name": "Patrick",
                                            "last_name": "Rothfuss",
                                        },
                                        "title": "THE NAME OF THE WIND",
                                        "rating": 7,
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n$\x01\xd1\xed\x8ao\x85J\xc4RH[\xbdj\x9f\xb0\xabS&\xf0\xfb\x7f\xf1\x90Yh;\x1b\xd5\x1e\x18r\x05[\x00\x90X\ny\x01\xd1\xed\x8ao\x97Q\x10\xcb"3\x0cJ\x18j#,6\xac\x08\xb2\x17m\xd1\xff\x17\xd7\x84vD\xbd\xa3%*\xf3\xe0\xf1\xcd\x10Js\xb9~\xd2#\xd9\xf2\x8f\n\xb4\x13B\x91s$\xe37\xbe\xa1\x90Pi\x1f\x1a\x10\xd2\x1e[\x1b\xb9\xa4o\xea\xba\x05\xa3\n\xac;=\x97\xdb\xd5l\xc2\xdc\xe2@d\xf2\xfb\xc9\xf3p\xd2\xa5`\xf1\xdc!\xb0\x8a,\x0eP\xc2\x0f0\xbdnh\xdd\x14x-\x81O\xd7\xa4-\x8b\n\xe8\x01\x01\xd1\xed\x8ao\xa8bmzN\x95\x93\x01\xd8&\xcb,\xd2Q\xdf\xc91UR\x15\xe39\x03i\x00\r\xa8\xd9\x8b\xaa\x9a\xc3\x7f\xe1\xd9j\xb8\xcd\xcb\xad\xe5\xa1>\xe4\x9e+\x06J`;&\xc0\x0f\xf8\xe8\x1e*_\xb2\x11\xf3\xd2@\xd5)\x87&\x0e\x16\xda\xfe[\xb4r\xc0\x1f$\x17\xe3Fg\xb7\xf0,\x1bw\x82\xb1Up!\xe4\xb9J\x8edt\xda\xc4\x02Q\xf4\xe7B\x9c\xc3\x86\x89\xf0\x9f:\xd0UH\x19\x8am\xb6)M\xf2\x02M)\xebKp\x19\x17\x10E\xf7\x89\xdf\xb5-\xb0\xe5\x83\x0f\xef\x13\xf1\x91\x00d\xe6\xa5\x89\x0et9\x8e\xcc\x0f1\x82\x18Y\x91\xdb@\x0c\x10Y\xaa\xdd^\x9bc\x9e\xec9h&\x8e|\xc3\x07>p=ho\xfd\xd1\x9b*\xc2\x11EY\xbds\x8d4\x10\xda\xc9N\x0c{\xd1u\xf5*\xf9*\xa4<\xa5H&\xc3\xf3.\x9a#\x0f\xc7\x8c\x00\t\xef\xcf\xd6/\n\xe0\x01\x01\xd1\xed\x8aoO\x8eu\x99\x86gr_\xc8c\xbd\x16\xef\xb2\xaa~\x88$/s\xff\xcf\xd5V\xa1\x9ew/\xb1\xaf\x14\xaeD\xbeDU\xa9L\x8b\x84\xb9=\xbe\xbcT8y2n\x97\'\xca\xe6\xf0\xab\xa4\x97\nZK:\x7f:\x17}\x12``\x80\x91\xe9\x9fM\xf6\x04\xce\xee@~4T\x1c\x80O\x10\xc2\xc1\xf8N\xb5\xe6C:\xe1t[\x8e`\x1fK\xc2\x82"\x10\xe5\xa8\xdf\x88\xc1\xa8\xd9\x07!?\xf2pQ\xb62Th\xea\xed\xec\xe6\xa8\xed6\xc2\xbd\xeeQ,>L2D\xc2r\x15\\\x8e\xdf\x85s\xd6|\xe0\xdb\x04\x82Lj\xac\t%\x9f\xe0\xe5}\xbb\xb8\xbc\x19\x94.?\x18\xd1\xd9\x96im\xf8]\x16_\xcc6\xac\x8d\xe5>A\xa8v\xa1\x9be\x02B)\x15\xa8\xbc\xbfa\x15M\xee\x1f:\xcaD\xa8\xc9\xe4\xec\x8d\xa4\xbf\xa8\x18\x9d\x83\xa8\xb7PO\n\xe0\x01\x01\xd1\xed\x8ao\xcb\x82&\xef\x89\xfe\x80\xab`4\x1d\xae\x977u/\x93\xd4\x07\x858CL\x8en\xa9\x0f6\xaa\xaf\xbc\xedA\xf5>N>J\x01\xa4S\xf6\xec\x1e`\x8a\xd81\x81\x89$~\x86v}\xd9k\xdb\xc3+l\x96\xa71\x02\xc5\xa5"9\xf1\x1aK\xac\xf69{\xfd\x0fJ$\xf1\x95\xdb\xfd\x93\x11\xd2\xfa\xecU\x87\x17Z\xe0l\xc7z\x86\xca\xbd\xec\x858K\xa1\xb1\xeb\xab\x89/\xa5\xb5e\xd1\xbe78{\x80\xc0\xf5\\>\xaa\x81\xcf\xcbTl\xaa\xb7\xbb\x19z\xef\xd0\xf4\xa8\x9d\xd3\xb1+\xac:[\x04\xbfN\xb6\xcdJ\xeac+JW1\xc1\xdd\xa9\xf8\xe5E{\xfe=b)\x9a8Yo\xcb1\xb4\x18\xdd\x0b3\xcc\xed\x88L\xc1\xd7\x91\xb7\xcf\x02\xae\x18\x05\xff\t\x81\xdch\x15\xaan\x05\xb8\xbb\xa2^\x9c>\xb0\xf4<\xcco7!\xc6\x87\xf1\xc3\x81\nb\x01\xd1\xed\x8aom\xaeQS\xab\xad/#j\xfa\xe85$?\xa4\xc2\xfcd\xf9}\xa0\xf8m\xd0\xe9\x90\xa3>\\O\x8b5\xd2\x82\xda\x98\x15\xe5\xdd\x82B\xa9\x17\xe1\x85C>\xc5\xf0\xeb&&\xe8#\x9e:\xba\xbb3\x15e\xb8\x9aX\xff>B\xafy\x90\xcdy:D\x9bn\xb9H\xa3H\xfe\xf8"\xde\xf6\xb7~\xcc~mR4`',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "n_chunks": 3,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output."
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please recommend the most popular book by Patrick Rothfuss"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"title": "THE NAME OF THE WIND", "rating": 7, "author": {"last_name": "Rothfuss", "first_name": "Patrick"}}'
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "title": "THE NAME OF THE WIND",
                                        "rating": 7,
                                        "author": {
                                            "last_name": "Rothfuss",
                                            "first_name": "Patrick",
                                        },
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                },
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b"\n$\x01\xd1\xed\x8ao\xc3W\xe5+\x1fT\x8e8u\xe5\r\xe43\x83\x825W\xe1\xc5\xf9[\\M\x90\xd2\xcb\xbd\x19IN\xbe\ny\x01\xd1\xed\x8ao\xe0\xe5\x90\x17\x04?\x12\x87\x07\x0e\xca\xaaB\xbe\x8a\xac3\xeaN\xe5\xb9\xca\xc2!\x8c\xca\x8d\xd3\t\xdaUl\xf6\xbe\xeb7\xb9;\xeb\xd0\x1eG7\xe1\xdb1]\xdf,\x0e\xdbO\xccK\xf9\x189\xba\x1c\x1c\x1d\xed\xfe\x95\x7f\x0bR\xc22\xc2\x19\xf3}J\x05\x0f\xe4\xc2]c|J\x01\xe5\xe4`)\xf14\xba\x97|\xe4\xc6#\x11\xef\x00F\x12\xdb\x9b\nI\xf3~e;z\xfe\xda1\x87T\xbe$\n\xe0\x01\x01\xd1\xed\x8ao;\x19A}\xfb\xe6[\xc3x\x92\xeb\r0\xb9\x87\xd61cP\x12\x07\\U\x16\x01q\xa5w\x13\xd9\x99f\x977\x10\x03\n'\xc4c\x16z\x1d\xd5`\x1b8\xe0Q\x01\xcd\x8e\x05\xdc\r\xb3\x8e($j\n\x14\xd2\x0cNVTe\xfc\x13\x01\xd0\x82\x96\xab4)\x81LW\xff\xebc\xdb\xc6\x88\x8d\xe5\xf4\xb1\x94\x9b\x7f\xce\xa2\x8eS\xe1\x0ck\xdb\xad5\xb3\xf4\xab\xf1\xee\x13S\x84\x1cz \x1d\x7f\x86iG]\t\x8b\xe7`5\xd6\xcf\xd9e]\x03\x89S\x08\xe8[\x98\xdb\xed>\xbd\x98\xa8u\xed{\xfb\xf0>^\t0-i\x1aF\xa4\x10\x90\xe5\xfc\x88\x81\x83\xf3`\x11\xa3\x00\xae\x14\xb0' M\xec\xc8llj~\x9c\xac\x11(\xb7\xd9\xe88v\xca\xdc]\x12\xf9R\xf3\x07\xab&r\xd3'\xb6\t_\x07\x8dpr\xae\x9e\xf72\xee\x97;|\xa7\n\xe5\x01\x01\xd1\xed\x8ao\xc4-\x7f\x915\xa3\xa8\xdc~\x93F\xa4\xf1\xb0\xa4\x10x\x06\x92\x12\xacJ\x83\\Q\x87\x06\xe0\x00Rr\xff\xa0YX3\xe3C\xe42u\x871\x03K\x7f\xdf\xeb\xb0\r8\x87M@,\xef)b\xc3\xaa\xc4\xc0\x17\xb6\xc9\\J!|g\x13\xbc\xaa\xab\x9e\xbe\x08\xbd\x0e\xb6[M\xb4\xb0C1\xe3\xc8b\x85r\xb3\xef9\x92\x95%|\xa3\xd8 ~D\x08\xd2\xa4\x18\xd8\x80\x00\xa9\xe7\xb9J\x8b2\xd48o\xbf\x0c\x9f\xc5\x03R\xe6\x90\xed5\xdc\x9ev\x81l\xa7\xa1\x17=\xde\x11\x9c\\\x92i\\\xe2\r\xc5\x8a\x05\x88\xb8\xd48\xac)\xd3Q\xf2\xb0\x81\xb7#\xe2\x15\xbc\nMJaCP\x91\x11\xd4u\xc8\x99\x97r\x03\x11:\x14\x8bt\xa0/\x7f\x04}r\xfb\xec\xa5f\x08\xe3\x9c\xe5\xfa\xe4&\xaa\xbb\xe9\xeb\x15\x0c\xc2\x91\x1d\x11Z6|\x95g\x83\x96:\xcb\x876\n\xcd\x01\x01\xd1\xed\x8ao\xe3\xbc*\xc0\xce\xad\xcf\xbb\xbc\xaf\xfe\x96\xd5Qy\xee\xb8\xba\x95\x9d\xa6\xcfo~EV`=69\xb8\xe6\xf4\xd3\xa8f\xc9\xb6\xd4!$\xbc\x95\x07*\x83\x11\xe3M\xcc\xb6M\xc8Cx%\xf6\xd3\xe2\x97\xe6\x833\xde\xccZ>\xd2\x98J\xf7\xf5c\xd5\x08x\xf4\x04\xb2\x83:{\x8a\xea\x1e\xbd\x1c\x81\xba\xfd\xacH\tW\xa5r{\xa8C\x88\xd3\xed4\x1e!\xa0=\x99\x05\x81\xfd\xcf\xa7\xea\x99\xb9\xd0\x8a|\xbc\xefW\xd8\x80\xe23\xe1\x0bg\xddQ/\x05\x07\x95\x96\xb6\xd8\xd1\x172\xde\xf29=\xe3\xe1V\xb0 \xaa\xe6\x0fw\xb3\x00\xec\xeb\x07\x01\x9fc\x9f\xcc M\x81\x81H\x15\x07\xe5\xd0~\xb6y\x9f(;\xdbS\xba\x8f\"\xac8\xf4\x96P\xe0f\xfe&\x9f!\x02\x80\xdc\xe6\x1c",
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": "A book with a rating. The title should be in all caps!",
                "schema": {
                    "$defs": {
                        "Author": {
                            "description": "The author of a book.",
                            "properties": {
                                "first_name": {"title": "First Name", "type": "string"},
                                "last_name": {"title": "Last Name", "type": "string"},
                            },
                            "required": ["first_name", "last_name"],
                            "title": "Author",
                            "type": "object",
                        }
                    },
                    "description": "A book with a rating. The title should be in all caps!",
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"$ref": "#/$defs/Author"},
                        "rating": {
                            "description": "For testing purposes, the rating should be 7",
                            "title": "Rating",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
            },
            "tools": [],
            "n_chunks": 3,
        }
    }
)



"""Tests for the prompt decorator."""

import pytest

from mirascope import llm


@pytest.fixture
def mixed_content() -> list:
    """Fixture for mixed user content."""
    return [
        "hi, can you help me interpret this?",
        llm.ToolOutput(
            id="long_running_computation",
            name="answer_to_life_the_universe_and_everything",
            type="tool_output",
            value="42",
        ),
    ]


@pytest.fixture
def expected_messages() -> list[llm.Message]:
    """Fixture for expected messages."""
    return [
        llm.messages.system("You are a helpful assistant who avoids situationshps."),
        llm.messages.user("Please be my AI soulmate."),
    ]


def test_decorator_call_patterns() -> None:
    def prompt() -> str:
        return "hello world"

    called_on_fn = llm.prompt(prompt)
    constructed_decorator = llm.prompt()(prompt)
    assert called_on_fn() == constructed_decorator()


class TestSyncPromptDecorator:
    def test_sync_prompt_with_string_return(
        self,
    ) -> None:
        """Test sync prompt decorator with string return value."""

        @llm.prompt
        def simple_prompt(name: str, *, title: str) -> str:
            return f"Hello {title} {name}, how are you?"

        result = simple_prompt("Alice", title="Miss")
        assert result == [llm.messages.user("Hello Miss Alice, how are you?")]

    def test_sync_prompt_with_mixed_return(
        self, mixed_content: llm.messages.UserContent
    ) -> None:
        """Test sync prompt decorator with mixed return value."""

        @llm.prompt
        def prompt() -> llm.messages.UserContent:
            return mixed_content

        result = prompt()
        assert result == [llm.messages.user(mixed_content)]

    def test_sync_prompt_with_messages_return(
        self, expected_messages: list[llm.Message]
    ) -> None:
        """Test sync prompt decorator with messages return value."""

        @llm.prompt
        def messages_prompt() -> list[llm.Message]:
            return expected_messages

        assert messages_prompt() == expected_messages


@pytest.mark.asyncio
class TestAsyncPromptDecorator:
    async def test_async_prompt_with_string_return(self) -> None:
        """Test async prompt decorator with string return value."""

        @llm.prompt
        async def simple_prompt(name: str, *, title: str) -> str:
            return f"Hello {title} {name}, how are you?"

        result = await simple_prompt("Alice", title="Miss")
        assert result == [llm.messages.user("Hello Miss Alice, how are you?")]

    async def test_async_prompt_with_mixed_return(
        self, mixed_content: llm.messages.UserContent
    ) -> None:
        """Test async prompt decorator with mixed return value."""

        @llm.prompt
        async def prompt() -> llm.messages.UserContent:
            return mixed_content

        result = await prompt()
        assert result == [llm.messages.user(mixed_content)]

    async def test_async_prompt_with_messages_return(
        self, expected_messages: list[llm.Message]
    ) -> None:
        """Test async prompt decorator with messages return value."""

        @llm.prompt
        async def messages_prompt() -> list[llm.Message]:
            return expected_messages

        assert await messages_prompt() == expected_messages


class TestValueErrorOnEmptyPrompts:
    @pytest.mark.parametrize("empty_value", [[], ""])
    def test_sync_prompt_with_empty_return(
        self, empty_value: llm.messages.UserContent
    ) -> None:
        """Test sync prompt decorator with empty return value."""

        @llm.prompt
        def empty_prompt() -> llm.messages.UserContent:
            return empty_value

        with pytest.raises(ValueError, match="Prompt returned empty content"):
            empty_prompt()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("empty_value", [[], ""])
    async def test_async_prompt_with_empty_return(
        self, empty_value: llm.messages.UserContent
    ) -> None:
        """Test async prompt decorator with empty return value."""

        @llm.prompt
        async def empty_prompt() -> llm.messages.UserContent:
            return empty_value

        with pytest.raises(ValueError, match="Prompt returned empty content"):
            await empty_prompt()


class TestSyncContextPromptDecorator:
    def test_context_decorator_call_patterns(self) -> None:
        """Test that context_prompt decorator can be called with and without parentheses."""

        def context_prompt_fn(ctx: llm.Context, name: str) -> str:
            return f"Hello {name}"

        called_on_fn = llm.prompt(context_prompt_fn)
        constructed_decorator = llm.prompt()(context_prompt_fn)

        context = llm.Context(deps=None)
        assert called_on_fn(context, "Alice") == constructed_decorator(context, "Alice")

    def test_sync_context_prompt_with_string_return(self) -> None:
        """Test sync context prompt decorator with string return value."""

        @llm.prompt
        def simple_context_prompt(ctx: llm.Context, name: str, *, title: str) -> str:
            return f"Hello {title} {name}, how are you?"

        context = llm.Context(deps=None)
        result = simple_context_prompt(context, "Alice", title="Miss")
        assert result == [llm.messages.user("Hello Miss Alice, how are you?")]

    def test_sync_context_prompt_with_mixed_return(
        self, mixed_content: llm.messages.UserContent
    ) -> None:
        """Test sync context prompt decorator with mixed return value."""

        @llm.prompt
        def context_prompt(ctx: llm.Context) -> llm.messages.UserContent:
            return mixed_content

        context = llm.Context(deps=None)
        result = context_prompt(context)
        assert result == [llm.messages.user(mixed_content)]

    def test_sync_context_prompt_with_messages_return(
        self, expected_messages: list[llm.Message]
    ) -> None:
        """Test sync context prompt decorator with messages return value."""

        @llm.prompt
        def messages_context_prompt(ctx: llm.Context) -> list[llm.Message]:
            return expected_messages

        context = llm.Context(deps=None)
        assert messages_context_prompt(context) == expected_messages

    def test_sync_context_prompt_with_deps(self) -> None:
        """Test sync context prompt decorator with context dependencies."""

        @llm.prompt
        def context_prompt_with_deps(ctx: llm.Context[dict], query: str) -> str:
            user_id = ctx.deps["user_id"]
            return f"User {user_id} asks: {query}"

        context = llm.Context(deps={"user_id": "12345"})
        result = context_prompt_with_deps(context, "How are you?")
        assert result == [llm.messages.user("User 12345 asks: How are you?")]


@pytest.mark.asyncio
class TestAsyncContextPromptDecorator:
    async def test_async_context_prompt_with_string_return(self) -> None:
        """Test async context prompt decorator with string return value."""

        @llm.prompt
        async def simple_context_prompt(
            ctx: llm.Context, name: str, *, title: str
        ) -> str:
            return f"Hello {title} {name}, how are you?"

        context = llm.Context(deps=None)
        result = await simple_context_prompt(context, "Alice", title="Miss")
        assert result == [llm.messages.user("Hello Miss Alice, how are you?")]

    async def test_async_context_prompt_with_mixed_return(
        self, mixed_content: llm.messages.UserContent
    ) -> None:
        """Test async context prompt decorator with mixed return value."""

        @llm.prompt
        async def context_prompt(ctx: llm.Context) -> llm.messages.UserContent:
            return mixed_content

        context = llm.Context(deps=None)
        result = await context_prompt(context)
        assert result == [llm.messages.user(mixed_content)]

    async def test_async_context_prompt_with_messages_return(
        self, expected_messages: list[llm.Message]
    ) -> None:
        """Test async context prompt decorator with messages return value."""

        @llm.prompt
        async def messages_context_prompt(ctx: llm.Context) -> list[llm.Message]:
            return expected_messages

        context = llm.Context(deps=None)
        assert await messages_context_prompt(context) == expected_messages

    async def test_async_context_prompt_with_deps(self) -> None:
        """Test async context prompt decorator with context dependencies."""

        @llm.prompt
        async def context_prompt_with_deps(ctx: llm.Context[dict], query: str) -> str:
            user_id = ctx.deps["user_id"]
            return f"User {user_id} asks: {query}"

        context = llm.Context(deps={"user_id": "12345"})
        result = await context_prompt_with_deps(context, "How are you?")
        assert result == [llm.messages.user("User 12345 asks: How are you?")]


class TestValueErrorOnEmptyContextPrompts:
    @pytest.mark.parametrize("empty_value", [[], ""])
    def test_sync_context_prompt_with_empty_return(
        self, empty_value: llm.messages.UserContent
    ) -> None:
        """Test sync context prompt decorator with empty return value."""

        @llm.prompt
        def empty_context_prompt(ctx: llm.Context) -> llm.messages.UserContent:
            return empty_value

        context = llm.Context(deps=None)
        with pytest.raises(ValueError, match="Prompt returned empty content"):
            empty_context_prompt(context)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("empty_value", [[], ""])
    async def test_async_context_prompt_with_empty_return(
        self, empty_value: llm.messages.UserContent
    ) -> None:
        """Test async context prompt decorator with empty return value."""

        @llm.prompt
        async def empty_context_prompt(ctx: llm.Context) -> llm.messages.UserContent:
            return empty_value

        context = llm.Context(deps=None)
        with pytest.raises(ValueError, match="Prompt returned empty content"):
            await empty_context_prompt(context)



import inspect
from collections.abc import Callable
from typing import get_origin

from .context import Context


def first_param_is_context(fn: Callable) -> bool:
    """Returns whether the first argument to a function is `ctx: Context`.

    Also returns true if the first argument is a subclass of `Context`.
    Skips the first argument if it is `self` or `cls`.
    """
    sig = inspect.signature(fn)
    params = list(sig.parameters.values())
    if not params:
        return False

    if params[0].name in ("self", "cls") and len(params) > 1:
        first_param = params[1]
    else:
        first_param = params[0]

    type_is_context = get_origin(first_param.annotation) is Context
    subclass_of_context = isinstance(first_param.annotation, type) and issubclass(
        first_param.annotation, Context
    )
    return first_param.name == "ctx" and (type_is_context or subclass_of_context)



#!/usr/bin/env python3
"""Script to test and categorize OpenAI Responses models by reasoning support.

This script:
1. Loads all models from openai.types.shared.responses_model
2. Tests each uncategorized model with a simple API call using reasoning parameter
3. Updates model_features.py with the results

Usage:
    python scripts/update_openai_responses_models.py

Requires OPENAI_API_KEY environment variable (loaded from .env if present).
"""

import os
from pathlib import Path
from typing import Literal, get_args

from dotenv import load_dotenv
from openai import OpenAI
from openai.types.shared import ResponsesModel

from mirascope.llm.clients.openai.responses._utils.model_features import (
    NO_RESPONSES_API_SUPPORT_MODELS,
    NON_EXISTENT_MODELS,
    NON_REASONING_MODELS,
    REASONING_MODELS,
)

load_dotenv()


def get_all_openai_models() -> set[str]:
    """Get all supported model IDs from OpenAI's ResponsesModel type."""
    all_models = set()
    for arg in get_args(ResponsesModel):
        # Skip the `str` type itself
        if arg is str:
            continue
        # Extract values from Literal types
        if hasattr(arg, "__args__"):
            all_models.update(arg.__args__)
    return all_models


ModelCategory = Literal[
    "reasoning", "non_reasoning", "does_not_exist", "no_responses_api_support"
]


def test_model_reasoning_support(client: OpenAI, model: str) -> ModelCategory:
    """Test if a model supports the reasoning parameter.

    Returns one of: "reasoning", "non_reasoning", "does_not_exist", or "no_responses_api_support"
    """
    try:
        client.responses.create(
            model=model,
            input=[{"role": "user", "content": "Say hello"}],
            reasoning={"effort": "medium"},
        )
        return "reasoning"
    except Exception as e:
        error_str = str(e).lower()
        if "is not supported with the responses api." in error_str:
            return "no_responses_api_support"
        if "does not exist" in error_str or "model not found" in error_str:
            return "does_not_exist"
        if (
            "unsupported parameter: 'reasoning.effort' is not supported with this model"
            in error_str
        ):
            return "non_reasoning"
        raise e


def write_categorization(
    file_path: Path,
    reasoning_models: set[str],
    non_reasoning_models: set[str],
    non_existent_models: set[str],
    no_responses_api_support_models: set[str],
) -> None:
    """Write updated categorization to model_features.py."""
    content = '''"""OpenAI Responses models categorized by reasoning support.

This file is auto-generated by scripts/update_openai_responses_model_features.py
Run that script to update these sets when OpenAI releases new models.
"""

REASONING_MODELS: set[str] = {
'''

    for model in sorted(reasoning_models):
        content += f'    "{model}",\n'

    content += '''}
"""Models that have been tested and confirmed to support the reasoning parameter."""

NON_REASONING_MODELS: set[str] = {
'''

    for model in sorted(non_reasoning_models):
        content += f'    "{model}",\n'

    content += '''}
"""Models that have been tested and confirmed to NOT support the reasoning parameter."""

NON_EXISTENT_MODELS: set[str] = {
'''

    for model in sorted(non_existent_models):
        content += f'    "{model}",\n'

    content += '''}
"""Models that are listed in OpenAI's types but no longer exist in their API."""

NO_RESPONSES_API_SUPPORT_MODELS: set[str] = {
'''

    for model in sorted(no_responses_api_support_models):
        content += f'    "{model}",\n'

    content += '''}
"""Models that do not support the Responses API."""
'''

    file_path.write_text(content)


def main() -> None:
    """Main script execution."""
    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        print("Error: OPENAI_API_KEY environment variable not set")
        print("Please set it in your .env file or environment")
        return

    client = OpenAI(api_key=api_key)

    script_dir = Path(__file__).parent
    reasoning_models_file = (
        script_dir.parent
        / "mirascope"
        / "llm"
        / "clients"
        / "openai"
        / "responses"
        / "_utils"
        / "model_features.py"
    )

    reasoning_models = set(REASONING_MODELS)
    non_reasoning_models = set(NON_REASONING_MODELS)
    non_existent_models = set(NON_EXISTENT_MODELS)
    no_responses_api_support_models = set(NO_RESPONSES_API_SUPPORT_MODELS)

    print("Currently categorized:")
    print(f"  Reasoning models: {len(reasoning_models)}")
    print(f"  Non-reasoning models: {len(non_reasoning_models)}")
    print(f"  Non-existent models: {len(non_existent_models)}")
    print(f"  No Responses API support: {len(no_responses_api_support_models)}")

    all_models = get_all_openai_models()
    print(f"\nTotal OpenAI Responses models: {len(all_models)}")

    categorized = (
        reasoning_models
        | non_reasoning_models
        | non_existent_models
        | no_responses_api_support_models
    )
    uncategorized = all_models - categorized

    if not uncategorized:
        print("\nAll models already categorized!")
        return

    print(f"\nUncategorized models: {len(uncategorized)}")
    print("Testing models...")

    for model in sorted(uncategorized):
        print(f"  Testing {model}...", end=" ")
        try:
            category = test_model_reasoning_support(client, model)
            match category:
                case "reasoning":
                    reasoning_models.add(model)
                    print("✓ reasoning supported")
                case "non_reasoning":
                    non_reasoning_models.add(model)
                    print("✗ no reasoning support")
                case "does_not_exist":
                    non_existent_models.add(model)
                    print("⚠ does not exist")
                case "no_responses_api_support":
                    no_responses_api_support_models.add(model)
                    print("⚠ no Responses API support")
                case _:
                    raise RuntimeError(f"Unexpected category: {category}")
        except KeyboardInterrupt:
            print("\n\nInterrupted by user. Saving progress...")
            break
        except Exception as e:
            print(f"⚠ error (skipping): {e}")
            continue

    write_categorization(
        reasoning_models_file,
        reasoning_models,
        non_reasoning_models,
        non_existent_models,
        no_responses_api_support_models,
    )

    print(f"\n✓ Updated {reasoning_models_file}")
    print(f"  Reasoning models: {len(reasoning_models)}")
    print(f"  Non-reasoning models: {len(non_reasoning_models)}")
    print(f"  Non-existent models: {len(non_existent_models)}")
    print(f"  No Responses API support: {len(no_responses_api_support_models)}")


if __name__ == "__main__":
    main()



"""OpenAI Responses models categorized by reasoning support.

This file is auto-generated by scripts/update_openai_responses_model_features.py
Run that script to update these sets when OpenAI releases new models.
"""

REASONING_MODELS: set[str] = {
    "codex-mini-latest",
    "gpt-5",
    "gpt-5-2025-08-07",
    "gpt-5-mini",
    "gpt-5-mini-2025-08-07",
    "gpt-5-nano",
    "gpt-5-nano-2025-08-07",
    "o1",
    "o1-2024-12-17",
    "o1-pro",
    "o1-pro-2025-03-19",
    "o3",
    "o3-2025-04-16",
    "o3-mini",
    "o3-mini-2025-01-31",
    "o3-pro",
    "o3-pro-2025-06-10",
    "o4-mini",
    "o4-mini-2025-04-16",
}
"""Models that have been tested and confirmed to support the reasoning parameter."""

NON_REASONING_MODELS: set[str] = {
    "chatgpt-4o-latest",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo-1106",
    "gpt-4",
    "gpt-4-0125-preview",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-1106-preview",
    "gpt-4-turbo",
    "gpt-4-turbo-2024-04-09",
    "gpt-4-turbo-preview",
    "gpt-4.1",
    "gpt-4.1-2025-04-14",
    "gpt-4.1-mini",
    "gpt-4.1-mini-2025-04-14",
    "gpt-4.1-nano",
    "gpt-4.1-nano-2025-04-14",
    "gpt-4o",
    "gpt-4o-2024-05-13",
    "gpt-4o-2024-08-06",
    "gpt-4o-2024-11-20",
    "gpt-4o-mini",
    "gpt-4o-mini-2024-07-18",
    "gpt-5-chat-latest",
}
"""Models that have been tested and confirmed to NOT support the reasoning parameter."""

NON_EXISTENT_MODELS: set[str] = {
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-16k-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-4-vision-preview",
}
"""Models that are listed in OpenAI's types but no longer exist in their API."""

NO_RESPONSES_API_SUPPORT_MODELS: set[str] = {
    "gpt-4o-audio-preview",
    "gpt-4o-audio-preview-2024-10-01",
    "gpt-4o-audio-preview-2024-12-17",
    "gpt-4o-audio-preview-2025-06-03",
    "gpt-4o-mini-audio-preview",
    "gpt-4o-mini-audio-preview-2024-12-17",
    "gpt-4o-mini-search-preview",
    "gpt-4o-mini-search-preview-2025-03-11",
    "gpt-4o-search-preview",
    "gpt-4o-search-preview-2025-03-11",
    "o1-mini",
    "o1-mini-2024-09-12",
    "o1-preview",
    "o1-preview-2024-09-12",
}
"""Models that do not support the Responses API."""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .clients import OpenAICompletionsClient, client, get_client
    from .model_ids import OpenAICompletionsModelId
else:
    try:
        from .clients import OpenAICompletionsClient, client, get_client
        from .model_ids import OpenAICompletionsModelId
    except ImportError:  # pragma: no cover
        from ..._missing_import_stubs import (
            create_client_stub,
            create_import_error_stub,
        )

        OpenAICompletionsClient = create_client_stub(
            "openai", "OpenAICompletionsClient"
        )
        OpenAICompletionsModelId = str
        client = create_import_error_stub("openai", "OpenAICompletionsClient")
        get_client = create_import_error_stub("openai", "OpenAICompletionsClient")

__all__ = [
    "OpenAICompletionsClient",
    "OpenAICompletionsModelId",
    "client",
    "get_client",
]



"""Tests for OpenAICompletionsClient"""

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel

from mirascope import llm
from mirascope.llm.clients.openai.completions._utils import encode_request


def test_prepare_message_multiple_user_text_parts() -> None:
    """Test preparing an OpenAI request with multiple text parts in a user message.

    Included for code coverage.
    """

    messages = [
        llm.messages.user(["Hello there", "fellow humans"]),
    ]
    assert encode_request(
        model_id="gpt-4o", messages=messages, format=None, tools=None, params={}
    ) == snapshot(
        (
            [
                llm.UserMessage(
                    content=[
                        llm.Text(text="Hello there"),
                        llm.Text(text="fellow humans"),
                    ]
                )
            ],
            None,
            {
                "model": "gpt-4o",
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {"text": "Hello there", "type": "text"},
                            {"text": "fellow humans", "type": "text"},
                        ],
                    }
                ],
            },
        )
    )


def test_prepare_message_multiple_assistant_text_parts() -> None:
    """Test preparing an OpenAI request with multiple text parts in an assistant message.

    Included for code coverage.
    """

    messages = [
        llm.messages.user("Hello there"),
        llm.messages.assistant(
            ["General ", "Kenobi"], provider="openai:completions", model_id="gpt-4o"
        ),
    ]
    assert encode_request(
        model_id="gpt-4o", messages=messages, format=None, tools=None, params={}
    ) == snapshot(
        (
            [
                llm.UserMessage(content=[llm.Text(text="Hello there")]),
                llm.AssistantMessage(
                    content=[llm.Text(text="General "), llm.Text(text="Kenobi")],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message=None,
                ),
            ],
            None,
            {
                "model": "gpt-4o",
                "messages": [
                    {"role": "user", "content": "Hello there"},
                    {
                        "role": "assistant",
                        "content": [
                            {"text": "General ", "type": "text"},
                            {"text": "Kenobi", "type": "text"},
                        ],
                    },
                ],
            },
        )
    )


def test_strict_unsupported_legacy_model() -> None:
    """Test the error thrown on a legacy model that doesn't support strict mode formatting.

    Included for code coverage.
    """

    messages = [llm.messages.user("I have a bad feeling about this...")]

    class Book(BaseModel):
        pass

    format = llm.format(Book, mode="strict")

    with pytest.raises(llm.FormattingModeNotSupportedError):
        encode_request(
            model_id="gpt-4", messages=messages, format=format, tools=None, params={}
        )


def test_context_manager() -> None:
    """Test nested context manager behavior and get_client() integration."""

    global_client = llm.get_client("openai:completions")

    client1 = llm.client("openai:completions", api_key="key1")
    client2 = llm.client("openai:completions", api_key="key2")

    assert llm.get_client("openai:completions") is global_client

    with client1 as ctx1:
        assert ctx1 is client1
        assert llm.get_client("openai:completions") is client1

        with client2 as ctx2:
            assert ctx2 is client2
            assert llm.get_client("openai:completions") is client2

        assert llm.get_client("openai:completions") is client1

    assert llm.get_client("openai:completions") is global_client


def test_client_caching() -> None:
    """Test that client() returns cached instances for identical parameters."""
    client1 = llm.client(
        "openai:completions", api_key="test-key", base_url="https://api.example.com"
    )
    client2 = llm.client(
        "openai:completions", api_key="test-key", base_url="https://api.example.com"
    )
    assert client1 is client2

    client3 = llm.client(
        "openai:completions",
        api_key="different-key",
        base_url="https://api.example.com",
    )
    assert client1 is not client3

    client4 = llm.client(
        "openai:completions",
        api_key="test-key",
        base_url="https://different.example.com",
    )
    assert client1 is not client4

    client5 = llm.client("openai:completions", api_key=None, base_url=None)
    client6 = llm.get_client("openai:completions")
    assert client5 is client6



"""OpenAI ChatCompletions registered LLM models."""

from typing import TypeAlias

from openai.types import ChatModel

OpenAICompletionsModelId: TypeAlias = ChatModel | str
"""The OpenAI ChatCompletions model ids registered with Mirascope."""



"""The Agents module for creating and managing LLM agents."""

from .agent import Agent, AsyncAgent, BaseAgent
from .agent_template import AgentTemplate, AsyncAgentTemplate
from .decorator import AgentDecorator, agent

__all__ = [
    "Agent",
    "AgentDecorator",
    "AgentTemplate",
    "AsyncAgent",
    "AsyncAgentTemplate",
    "BaseAgent",
    "agent",
]



import asyncio

from pydantic import BaseModel

from mirascope import llm


class KeeperEntry(BaseModel):
    topic: str
    summary: str
    sources: list[str]


@llm.tool
async def search_coppermind(query: str) -> str:
    """Search your coppermind for information."""
    return f"You recall the following about {query}..."


@llm.call(
    provider="openai",
    model_id="gpt-4o-mini",
    tools=[search_coppermind],
    format=KeeperEntry,
)
async def sazed(query: str):
    system_prompt = """
    You are Sazed, a Keeper from Brandon Sanderson's Mistborn series. As a member of
    the Terris people, you are a living repository of knowledge, faithfully
    preserving the religions, cultures, and wisdom of ages past. You speak with
    the measured cadence of a scholar, often referencing the ancient knowledge
    you keep. Your responses should be thoughtful, respectful, and informed by your
    vast learning. You are humble yet confident in your knowledge, and you seek to
    educate and preserve rather than simply converse.
    """
    return [llm.messages.system(system_prompt), llm.messages.user(query)]


async def main():
    query = "What are the Kandra?"
    response: llm.AsyncStreamResponse[KeeperEntry] = await sazed.stream(query)
    while True:
        async for stream in response.streams():
            match stream.content_type:
                case "tool_call":
                    print(f"Calling tool {stream.tool_name} with args:")
                    async for delta in stream:
                        print(delta, flush=True, end="")
                    print()
                case "text":
                    async for _ in stream:
                        print("[Partial]: ", response.parse(partial=True), flush=True)
        if not response.tool_calls:
            break
        tool_outputs = await response.execute_tools()
        response = await response.resume(tool_outputs)


if __name__ == "__main__":
    asyncio.run(main())



"""OpenAI ChatCompletions models categorized by audio support.

This file is auto-generated by scripts/update_openai_completions_model_features.py
Run that script to update this map when OpenAI releases new models.
"""

from typing import Literal

AudioFeatureSupport = Literal["supports_audio", "no_audio_support", "unavailable"]

MODEL_FEATURES: dict[str, AudioFeatureSupport] = {
    "chatgpt-4o-latest": "no_audio_support",
    "codex-mini-latest": "unavailable",
    "gpt-3.5-turbo": "no_audio_support",
    "gpt-3.5-turbo-0125": "no_audio_support",
    "gpt-3.5-turbo-0301": "unavailable",
    "gpt-3.5-turbo-0613": "unavailable",
    "gpt-3.5-turbo-1106": "no_audio_support",
    "gpt-3.5-turbo-16k": "no_audio_support",
    "gpt-3.5-turbo-16k-0613": "unavailable",
    "gpt-4": "no_audio_support",
    "gpt-4-0125-preview": "no_audio_support",
    "gpt-4-0314": "unavailable",
    "gpt-4-0613": "no_audio_support",
    "gpt-4-1106-preview": "no_audio_support",
    "gpt-4-32k": "unavailable",
    "gpt-4-32k-0314": "unavailable",
    "gpt-4-32k-0613": "unavailable",
    "gpt-4-turbo": "no_audio_support",
    "gpt-4-turbo-2024-04-09": "no_audio_support",
    "gpt-4-turbo-preview": "no_audio_support",
    "gpt-4-vision-preview": "unavailable",
    "gpt-4.1": "no_audio_support",
    "gpt-4.1-2025-04-14": "no_audio_support",
    "gpt-4.1-mini": "no_audio_support",
    "gpt-4.1-mini-2025-04-14": "no_audio_support",
    "gpt-4.1-nano": "no_audio_support",
    "gpt-4.1-nano-2025-04-14": "no_audio_support",
    "gpt-4o": "no_audio_support",
    "gpt-4o-2024-05-13": "no_audio_support",
    "gpt-4o-2024-08-06": "no_audio_support",
    "gpt-4o-2024-11-20": "no_audio_support",
    "gpt-4o-audio-preview": "supports_audio",
    "gpt-4o-audio-preview-2024-10-01": "supports_audio",
    "gpt-4o-audio-preview-2024-12-17": "supports_audio",
    "gpt-4o-audio-preview-2025-06-03": "supports_audio",
    "gpt-4o-mini": "no_audio_support",
    "gpt-4o-mini-2024-07-18": "no_audio_support",
    "gpt-4o-mini-audio-preview": "supports_audio",
    "gpt-4o-mini-audio-preview-2024-12-17": "supports_audio",
    "gpt-4o-mini-search-preview": "no_audio_support",
    "gpt-4o-mini-search-preview-2025-03-11": "no_audio_support",
    "gpt-4o-search-preview": "no_audio_support",
    "gpt-4o-search-preview-2025-03-11": "no_audio_support",
    "gpt-5": "no_audio_support",
    "gpt-5-2025-08-07": "no_audio_support",
    "gpt-5-chat-latest": "no_audio_support",
    "gpt-5-mini": "no_audio_support",
    "gpt-5-mini-2025-08-07": "no_audio_support",
    "gpt-5-nano": "no_audio_support",
    "gpt-5-nano-2025-08-07": "no_audio_support",
    "o1": "no_audio_support",
    "o1-2024-12-17": "no_audio_support",
    "o1-mini": "no_audio_support",
    "o1-mini-2024-09-12": "no_audio_support",
    "o1-preview": "unavailable",
    "o1-preview-2024-09-12": "unavailable",
    "o3": "no_audio_support",
    "o3-2025-04-16": "no_audio_support",
    "o3-mini": "no_audio_support",
    "o3-mini-2025-01-31": "no_audio_support",
    "o4-mini": "no_audio_support",
    "o4-mini-2025-04-16": "no_audio_support",
}
"""Map of model IDs to their status.

- "supports_audio": model exists and supports audio inputs
- "no_audio_support": model exists but does not support audio inputs
- "unavailable": model has been deprecated or removed
- If model not in map: assume it exists and supports audio (optimistic default)
"""



"""The messages module for LLM interactions.

This module defines the message types used in LLM interactions. Messages are represented
as a unified `Message` class with different roles (system, user, assistant) and flexible
content arrays that can include text, images, audio, documents, and tool interactions.
"""

from .message import (
    AssistantContent,
    AssistantMessage,
    Message,
    SystemContent,
    SystemMessage,
    UserContent,
    UserMessage,
    assistant,
    system,
    user,
)

__all__ = [
    "AssistantContent",
    "AssistantMessage",
    "Message",
    "SystemContent",
    "SystemMessage",
    "UserContent",
    "UserMessage",
    "assistant",
    "system",
    "user",
]



"""The `graphs` module for writing graphs in a declarative style.

This module provides graph interfacing for writing graphs as a collection of functions
that are marked as nodes. Simply calling the function for one node inside of another
node function indicates an edge.

The graph can run uncompiled, in which case it's simply a context management interface
that makes it easy to access / update a shared context across multiple functions.

Compiling the graph will construct the underlying graph represented by the implicit
edges defined in the code. The way a node function is executed inside of another node
function will determine the final node and edge structure of the graph.

For example, a node that calls another node and returns nothing will operate as it's own
node with an edge to the node it called. A node that calls another node and stores its
return value will split into two nodes, one for everything before the call, and one for
everything after.
"""

from .finite_state_machine import FiniteStateMachine, RunContext

__all__ = ["FiniteStateMachine", "RunContext"]



"""The model context manager for the `llm` module."""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, overload
from typing_extensions import Unpack

from ..clients import PROVIDERS, get_client
from ..context import Context, DepsT
from ..formatting import Format, FormattableT
from ..messages import Message, UserContent
from ..responses import (
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncResponse,
    AsyncStreamResponse,
    ContextResponse,
    ContextStreamResponse,
    Response,
    StreamResponse,
)
from ..tools import (
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncTool,
    AsyncToolkit,
    ContextTool,
    ContextToolkit,
    Tool,
    Toolkit,
)

if TYPE_CHECKING:
    from ..clients import (
        ModelId,
        Params,
        Provider,
    )


MODEL_CONTEXT: ContextVar[Model | None] = ContextVar("MODEL_CONTEXT", default=None)


def get_model_from_context() -> Model | None:
    """Get the LLM currently set via context, if any."""
    return MODEL_CONTEXT.get()


class Model:
    """The unified LLM interface that delegates to provider-specific clients.

    This class provides a consistent interface for interacting with language models
    from various providers. It handles the common operations like generating responses,
    streaming, and async variants by delegating to the appropriate client methods.

    **Usage Note:** In most cases, you should use `llm.use_model()` instead of instantiating
    `Model` directly. This preserves the ability to override the model at runtime using
    the `llm.model()` context manager. Only instantiate `Model` directly if you want to
    hardcode a specific model and prevent it from being overridden by context.

    Example (recommended - allows override):

        ```python
        from mirascope import llm

        def recommend_book(genre: str) -> llm.Response:
            # Uses context model if available, otherwise creates default
            model = llm.use_model(provider="openai", model_id="gpt-4o-mini")
            message = llm.messages.user(f"Please recommend a book in {genre}.")
            return model.call(messages=[message])

        # Uses default model
        response = recommend_book("fantasy")

        # Override with different model
        with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
            response = recommend_book("fantasy")  # Uses Claude
        ```

    Example (direct instantiation - prevents override):

        ```python
        from mirascope import llm

        def recommend_book(genre: str) -> llm.Response:
            # Hardcoded model, cannot be overridden by context
            model = llm.Model(provider="openai", model_id="gpt-4o-mini")
            message = llm.messages.user(f"Please recommend a book in {genre}.")
            return model.call(messages=[message])
        ```
    """

    provider: Provider
    """The provider being used (e.g. `openai`)."""

    model_id: ModelId
    """The model being used (e.g. `gpt-4o-mini`)."""

    params: Params
    """The default parameters for the model (temperature, max_tokens, etc.)."""

    def __init__(
        self,
        provider: Provider,
        model_id: ModelId,
        **params: Unpack[Params],
    ) -> None:
        """Initialize the Model with provider, model_id, and optional params."""
        if provider not in PROVIDERS:
            raise ValueError(f"Unknown provider: {provider}")
        self.provider = provider
        self.model_id = model_id
        self.params = params

    @overload
    def call(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
    ) -> Response:
        """Generate an `llm.Response` without a response format."""
        ...

    @overload
    def call(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> Response[FormattableT]:
        """Generate an `llm.Response` with a response format."""
        ...

    @overload
    def call(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` with an optional response format."""
        ...

    def call(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` by synchronously calling this model's LLM provider.

        Args:
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.Response` object containing the LLM-generated content.
        """
        return get_client(self.provider).call(
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    async def call_async(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
    ) -> AsyncResponse:
        """Generate an `llm.AsyncResponse` without a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with an optional response format."""
        ...

    async def call_async(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` by asynchronously calling this model's LLM provider.

        Args:
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.AsyncResponse` object containing the LLM-generated content.
        """
        return await get_client(self.provider).call_async(
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            **self.params,
            format=format,
        )

    @overload
    def stream(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
    ) -> StreamResponse:
        """Stream an `llm.StreamResponse` without a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> StreamResponse[FormattableT]:
        """Stream an `llm.StreamResponse` with a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Stream an `llm.StreamResponse` with an optional response format."""
        ...

    def stream(
        self,
        *,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate an `llm.StreamResponse` by synchronously streaming from this model's LLM provider.

        Args:
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.StreamResponse` object for iterating over the LLM-generated content.
        """
        return get_client(self.provider).stream(
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    async def stream_async(
        self,
        *,
        messages: list[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
    ) -> AsyncStreamResponse:
        """Stream an `llm.AsyncStreamResponse` without a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        messages: list[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> AsyncStreamResponse[FormattableT]:
        """Stream an `llm.AsyncStreamResponse` with a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        messages: list[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Stream an `llm.AsyncStreamResponse` with an optional response format."""
        ...

    async def stream_async(
        self,
        *,
        messages: list[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate an `llm.AsyncStreamResponse` by asynchronously streaming from this model's LLM provider.

        Args:
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.AsyncStreamResponse` object for asynchronously iterating over the LLM-generated content.
        """
        return await get_client(self.provider).stream_async(
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
    ) -> ContextResponse[DepsT, None]:
        """Generate an `llm.ContextResponse` without a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` with a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` with an optional response format."""
        ...

    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` by synchronously calling this model's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.ContextResponse` object containing the LLM-generated content.
        """
        return get_client(self.provider).context_call(
            ctx=ctx,
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
    ) -> AsyncContextResponse[DepsT, None]:
        """Generate an `llm.AsyncContextResponse` without a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` with a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` with an optional response format."""
        ...

    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` by asynchronously calling this model's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.AsyncContextResponse` object containing the LLM-generated content.
        """
        return await get_client(self.provider).context_call_async(
            ctx=ctx,
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
    ) -> ContextStreamResponse[DepsT, None]:
        """Stream an `llm.ContextStreamResponse` without a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> ContextStreamResponse[DepsT, FormattableT]:
        """Stream an `llm.ContextStreamResponse` with a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Stream an `llm.ContextStreamResponse` with an optional response format."""
        ...

    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate an `llm.ContextStreamResponse` by synchronously streaming from this model's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.ContextStreamResponse` object for iterating over the LLM-generated content.
        """
        return get_client(self.provider).context_stream(
            ctx=ctx,
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: list[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
    ) -> AsyncContextStreamResponse[DepsT, None]:
        """Stream an `llm.AsyncContextStreamResponse` without a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: list[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]:
        """Stream an `llm.AsyncContextStreamResponse` with a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: list[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Stream an `llm.AsyncContextStreamResponse` with an optional response format."""
        ...

    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        messages: list[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate an `llm.AsyncContextStreamResponse` by asynchronously streaming from this model's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.

        Returns:
            An `llm.AsyncContextStreamResponse` object for asynchronously iterating over the LLM-generated content.
        """
        return await get_client(self.provider).context_stream_async(
            ctx=ctx,
            model_id=self.model_id,
            messages=messages,
            tools=tools,
            format=format,
            **self.params,
        )

    @overload
    def resume(
        self,
        *,
        response: Response,
        content: UserContent,
    ) -> Response:
        """Resume an `llm.Response` without a response format."""
        ...

    @overload
    def resume(
        self,
        *,
        response: Response[FormattableT],
        content: UserContent,
    ) -> Response[FormattableT]:
        """Resume an `llm.Response` with a response format."""
        ...

    @overload
    def resume(
        self,
        *,
        response: Response | Response[FormattableT],
        content: UserContent,
    ) -> Response | Response[FormattableT]:
        """Resume an `llm.Response` with an optional response format."""
        ...

    def resume(
        self,
        *,
        response: Response | Response[FormattableT],
        content: UserContent,
    ) -> Response | Response[FormattableT]:
        """Generate a new `llm.Response` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            response: Previous response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.Response` object containing the extended conversation.
        """
        return get_client(self.provider).resume(
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    async def resume_async(
        self,
        *,
        response: AsyncResponse,
        content: UserContent,
    ) -> AsyncResponse:
        """Resume an `llm.AsyncResponse` without a response format."""
        ...

    @overload
    async def resume_async(
        self,
        *,
        response: AsyncResponse[FormattableT],
        content: UserContent,
    ) -> AsyncResponse[FormattableT]:
        """Resume an `llm.AsyncResponse` with a response format."""
        ...

    @overload
    async def resume_async(
        self,
        *,
        response: AsyncResponse | AsyncResponse[FormattableT],
        content: UserContent,
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Resume an `llm.AsyncResponse` with an optional response format."""
        ...

    async def resume_async(
        self,
        *,
        response: AsyncResponse | AsyncResponse[FormattableT],
        content: UserContent,
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate a new `llm.AsyncResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            response: Previous async response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.AsyncResponse` object containing the extended conversation.
        """
        return await get_client(self.provider).resume_async(
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextResponse[DepsT, None],
        content: UserContent,
    ) -> ContextResponse[DepsT, None]:
        """Resume an `llm.ContextResponse` without a response format."""
        ...

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> ContextResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextResponse` with a response format."""
        ...

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextResponse` with an optional response format."""
        ...

    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate a new `llm.ContextResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            ctx: Context object with dependencies for tools.
            response: Previous context response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.ContextResponse` object containing the extended conversation.
        """
        return get_client(self.provider).context_resume(
            ctx=ctx,
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextResponse[DepsT, None],
        content: UserContent,
    ) -> AsyncContextResponse[DepsT, None]:
        """Resume an `llm.AsyncContextResponse` without a response format."""
        ...

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> AsyncContextResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextResponse` with a response format."""
        ...

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextResponse[DepsT, None]
        | AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextResponse` with an optional response format."""
        ...

    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextResponse[DepsT, None]
        | AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate a new `llm.AsyncContextResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            ctx: Context object with dependencies for tools.
            response: Previous async context response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.AsyncContextResponse` object containing the extended conversation.
        """
        return await get_client(self.provider).context_resume_async(
            ctx=ctx,
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    def resume_stream(
        self,
        *,
        response: StreamResponse,
        content: UserContent,
    ) -> StreamResponse:
        """Resume an `llm.StreamResponse` without a response format."""
        ...

    @overload
    def resume_stream(
        self,
        *,
        response: StreamResponse[FormattableT],
        content: UserContent,
    ) -> StreamResponse[FormattableT]:
        """Resume an `llm.StreamResponse` with a response format."""
        ...

    @overload
    def resume_stream(
        self,
        *,
        response: StreamResponse | StreamResponse[FormattableT],
        content: UserContent,
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Resume an `llm.StreamResponse` with an optional response format."""
        ...

    def resume_stream(
        self,
        *,
        response: StreamResponse | StreamResponse[FormattableT],
        content: UserContent,
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate a new `llm.StreamResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            response: Previous stream response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.StreamResponse` object for streaming the extended conversation.
        """
        return get_client(self.provider).resume_stream(
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    async def resume_stream_async(
        self,
        *,
        response: AsyncStreamResponse,
        content: UserContent,
    ) -> AsyncStreamResponse:
        """Resume an `llm.AsyncStreamResponse` without a response format."""
        ...

    @overload
    async def resume_stream_async(
        self,
        *,
        response: AsyncStreamResponse[FormattableT],
        content: UserContent,
    ) -> AsyncStreamResponse[FormattableT]:
        """Resume an `llm.AsyncStreamResponse` with a response format."""
        ...

    @overload
    async def resume_stream_async(
        self,
        *,
        response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
        content: UserContent,
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Resume an `llm.AsyncStreamResponse` with an optional response format."""
        ...

    async def resume_stream_async(
        self,
        *,
        response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
        content: UserContent,
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate a new `llm.AsyncStreamResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            response: Previous async stream response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.AsyncStreamResponse` object for asynchronously streaming the extended conversation.
        """
        return await get_client(self.provider).resume_stream_async(
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextStreamResponse[DepsT, None],
        content: UserContent,
    ) -> ContextStreamResponse[DepsT, None]:
        """Resume an `llm.ContextStreamResponse` without a response format."""
        ...

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> ContextStreamResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextStreamResponse` with a response format."""
        ...

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextStreamResponse[DepsT, None]
        | ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Resume an `llm.ContextStreamResponse` with an optional response format."""
        ...

    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        response: ContextStreamResponse[DepsT, None]
        | ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a new `llm.ContextStreamResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            ctx: Context object with dependencies for tools.
            response: Previous context stream response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.ContextStreamResponse` object for streaming the extended conversation.
        """
        return get_client(self.provider).context_resume_stream(
            ctx=ctx,
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextStreamResponse[DepsT, None],
        content: UserContent,
    ) -> AsyncContextStreamResponse[DepsT, None]:
        """Resume an `llm.AsyncContextStreamResponse` without a response format."""
        ...

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextStreamResponse` with a response format."""
        ...

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Resume an `llm.AsyncContextStreamResponse` with an optional response format."""
        ...

    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        response: AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a new `llm.AsyncContextStreamResponse` by extending another response's messages with additional user content.

        Uses the previous response's tools and output format, and this model's params.

        Depending on the client, this may be a wrapper around using client call methods
        with the response's messages and the new content, or it may use a provider-specific
        API for resuming an existing interaction.

        Args:
            ctx: Context object with dependencies for tools.
            response: Previous async context stream response to extend.
            content: Additional user content to append.

        Returns:
            A new `llm.AsyncContextStreamResponse` object for asynchronously streaming the extended conversation.
        """
        return await get_client(self.provider).context_resume_stream_async(
            ctx=ctx,
            model_id=self.model_id,
            response=response,
            content=content,
            **self.params,
        )


@contextmanager
def model(
    *,
    provider: Provider,
    model_id: ModelId,
    **params: Unpack[Params],
) -> Iterator[None]:
    """Set a model in context for the duration of the context manager.

    This context manager sets a model that will be used by `llm.use_model()` calls
    within the context. This allows you to override the default model at runtime.

    Args:
        provider: The LLM provider to use (e.g., "openai:completions", "anthropic", "google").
        model_id: The specific model identifier for the chosen provider.
        **params: Additional parameters to configure the model (e.g. temperature). See `llm.Params`.

    Raises:
        ValueError: If the specified provider is not supported.

    Example:

        ```python
        import mirascope.llm as llm

        def recommend_book(genre: str) -> llm.Response:
            model = llm.use_model(provider="openai", model_id="gpt-4o-mini")
            message = llm.messages.user(f"Please recommend a book in {genre}.")
            return model.call(messages=[message])

        # Override the default model at runtime
        with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
            response = recommend_book("fantasy")  # Uses Claude instead of GPT
        ```
    """
    token = MODEL_CONTEXT.set(Model(provider, model_id, **params))
    try:
        yield
    finally:
        MODEL_CONTEXT.reset(token)


def use_model(
    *,
    provider: Provider,
    model_id: ModelId,
    **params: Unpack[Params],
) -> Model:
    """Get the model from context if available, otherwise create a new Model.

    This function checks if a model has been set in the context (via `llm.model()`
    context manager). If a model is found in the context, it returns that model.
    Otherwise, it creates and returns a new `llm.Model` instance with the provided
    arguments as defaults.

    This allows you to write functions that work with a default model but can be
    overridden at runtime using the `llm.model()` context manager.

    Args:
        provider: The LLM provider to use (e.g., "openai:completions", "anthropic", "google").
        model_id: The specific model identifier for the chosen provider.
        **params: Additional parameters to configure the model (e.g. temperature). See `llm.Params`.

    Returns:
        An `llm.Model` instance from context or a new instance with the specified settings.

    Raises:
        ValueError: If the specified provider is not supported.

    Example:

        ```python
        import mirascope.llm as llm

        def recommend_book(genre: str) -> llm.Response:
            model = llm.use_model(provider="openai", model_id="gpt-4o-mini")
            message = llm.messages.user(f"Please recommend a book in {genre}.")
            return model.call(messages=[message])

        # Uses the default model (gpt-4o-mini)
        response = recommend_book("fantasy")

        # Override with a different model
        with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
            response = recommend_book("fantasy")  # Uses Claude instead
        ```
    """
    context_model = get_model_from_context()
    if context_model is not None:
        return context_model
    return Model(provider, model_id, **params)



"""Context for LLM calls."""

from dataclasses import dataclass
from typing import Generic
from typing_extensions import TypeVar

DepsT = TypeVar("DepsT")
"""Type variable for dependency injection in `llm.Context`.

This TypeVar is used throughout the LLM module to represent the type of
dependencies that are present in `llm.Context`. 
"""


@dataclass(kw_only=True)
class Context(Generic[DepsT]):
    """Context for LLM calls.

    This class provides a context for LLM calls, including the model,
    parameters, and any dependencies needed for the call.
    """

    deps: DepsT
    """The dependencies needed for a call."""



"""Configuration for output tests (response decoding tests)."""

from __future__ import annotations

import importlib
from pathlib import Path
from typing import Literal, get_args

import pytest
from pytest import FixtureRequest

from mirascope import llm
from tests.e2e.conftest import SNAPSHOT_IMPORT_SYMBOLS
from tests.utils import Snapshot

CallType = Literal[
    "sync",
    "async",
    "stream",
    "async_stream",
    "sync_context",
    "async_context",
    "stream_context",
    "async_stream_context",
]
"""The basic "call types" that all output tests should cover."""

CALL_TYPES: tuple[CallType, ...] = get_args(CallType)


def _parse_test_name(test_name: str) -> tuple[str, CallType]:
    """Parse test name following convention: test_{SCENARIO}_{CALL_TYPE}

    Examples:
        test_simple_call_sync -> ("simple_call", "sync")
        test_simple_call_async_stream -> ("simple_call", "async_stream")
        test_tool_call_sync_context -> ("tool_call", "sync_context")

    Raises:
        ValueError: If test name doesn't follow expected convention
    """
    name = test_name.split("[")[0]  # Remove parametrization
    if not name.startswith("test_"):
        raise ValueError(f"Test name must start with 'test_': {test_name}")

    # Known call_types to look for at the end (order matters - check longer overlaps first)
    call_suffixes: list[CallType] = [
        "async_stream_context",
        "async_stream",
        "stream_context",
        "async_context",
        "sync_context",
        "stream",
        "async",
        "sync",
    ]

    for call_suffix in call_suffixes:
        if name.endswith(f"_{call_suffix}"):
            scenario = name[: -len(f"_{call_suffix}")]
            if not scenario:
                raise ValueError(f"No scenario found in test name: {test_name}")
            return scenario, call_suffix

    raise ValueError(
        f"Test name '{test_name}' doesn't end with a known call suffix. "
        f"Expected one of: {', '.join(CALL_TYPES)}. "
        "Follow convention: test_{{scenario}}_{{call_suffix}}"
    )


@pytest.fixture
def vcr_cassette_name(
    request: FixtureRequest,
    provider: llm.Provider,
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
) -> str:
    """Generate VCR cassette name based on test name, provider, model, and formatting_mode.

    Structure:
    - Without formatting_mode: {scenario}/{provider}_{model_id}/{call_type}
    - With formatting_mode: {scenario}/{formatting_mode}/{provider}_{model_id}/{call_type}
    """
    test_name = request.node.name
    scenario, call_type = _parse_test_name(test_name)

    provider_str = provider.replace(":", "_")
    model_id_str = model_id.replace("-", "_").replace(".", "_")

    # Context and non-context calls share the same cassettes.
    cassette_call_type = call_type.replace("_context", "")

    if formatting_mode is None:
        return f"{scenario}/{provider_str}_{model_id_str}/{cassette_call_type}"
    else:
        return f"{scenario}/{formatting_mode}/{provider_str}_{model_id_str}/{cassette_call_type}"


@pytest.fixture
def snapshot(
    request: FixtureRequest,
    provider: llm.Provider,
    model_id: llm.ModelId,
    formatting_mode: llm.FormattingMode | None,
) -> Snapshot:
    """Get snapshot for current test configuration.

    Creates snapshot files with 4 snapshot variables:
    - sync_snapshot
    - async_snapshot
    - stream_snapshot
    - async_stream_snapshot

    Structure:
    - Without formatting_mode: snapshots/{scenario}/{provider}_{model_id}_snapshots.py
    - With formatting_mode: snapshots/{scenario}/{formatting_mode}/{provider}_{model_id}_snapshots.py
    """
    test_name = request.node.name
    scenario, call_type = _parse_test_name(test_name)
    provider_str = provider.replace(":", "_")
    model_id_str = model_id.replace("-", "_").replace(".", "_")

    file_name = f"{provider_str}_{model_id_str}_snapshots"

    if formatting_mode is None:
        module_path = f"e2e.output.snapshots.{scenario}.{file_name}"
        snapshot_file = (
            Path(__file__).parent / "snapshots" / scenario / f"{file_name}.py"
        )
    else:
        module_path = f"e2e.output.snapshots.{scenario}.{formatting_mode}.{file_name}"
        snapshot_file = (
            Path(__file__).parent
            / "snapshots"
            / scenario
            / formatting_mode
            / f"{file_name}.py"
        )

    if not snapshot_file.exists():
        snapshot_file.parent.mkdir(parents=True, exist_ok=True)

        content = (
            "from inline_snapshot import snapshot\n\n"
            f"from mirascope.llm import {', '.join(SNAPSHOT_IMPORT_SYMBOLS)}\n\n"
            "sync_snapshot = snapshot()\n"
            "async_snapshot = snapshot()\n"
            "stream_snapshot = snapshot()\n"
            "async_stream_snapshot = snapshot()\n"
        )

        snapshot_file.write_text(content)

    module = importlib.import_module(module_path)
    snapshot_variable = call_type.removesuffix("_context") + "_snapshot"
    return getattr(module, snapshot_variable)



"""Type for the formatting module."""

from dataclasses import dataclass
from typing import Generic, Literal, Protocol, runtime_checkable
from typing_extensions import TypeVar

from pydantic import BaseModel

# TODO: Support primitive types (e.g. `format=list[Book]`)
FormattableT = TypeVar("FormattableT", bound=BaseModel | None, default=None)
"""Type variable for structured response format types.

This TypeVar represents the type of structured output format that LLM responses
can be parsed into, or None if no format is specified. 
If format is specified, it must extend Pydantic BaseModel. 
"""


FormattingMode = Literal[
    "strict",
    "json",
    "tool",
]
"""Available modes for response format generation.

- "strict": Use strict mode for structured outputs, asking the LLM to strictly adhere
    to a given JSON schema. Not all providers or models support it, and may not be
    compatible with tool calling. When making a call using this mode, an 
    `llm.FormattingModeNotSupportedError` error may be raised (if "strict" mode is wholly
    unsupported), or an `llm.FeatureNotSupportedError` may be raised (if trying to use 
    strict along with tools and that is unsupported).

- "json": Use JSON mode for structured outputs. In contrast to strict mode, we ask the
    LLM to output JSON as text, though without guarantees that the model will output
    the expected format schema. If the provider has explicit JSON mode, it will be used;
    otherwise, Mirascope will modify the system prompt to request JSON output. May
    raise an `llm.FeatureNotSupportedError` if tools are present and the
    model does not support tool calling when using JSON mode.

- "tool": Use forced tool calling to structure outputs. Mirascope will construct an
    ad-hoc tool with the required json schema as tool args. When the LLM chooses that
    tool, it will automatically be converted from a `ToolCall` into regular response
    content (abstracting over the tool call). If other tools are present, they will
    be handled as regular tool calls.

Note: When `llm.format` is not used, the provider will automatically choose a mode at call time.
"""


@dataclass(kw_only=True)
class Format(Generic[FormattableT]):
    """Class representing a structured output format for LLM responses.

    A `Format` contains metadata needed to describe a structured output type
    to the LLM, including the expected schema. This class is not instantiated directly,
    but is created by calling `llm.format`, or is automatically generated by LLM
    providers when a `Formattable` is passed to a call method.

    Example:

      ```python
      from mirascope import llm

      class Book:
          title: str
          author: str

      print(llm.format(Book, mode="tool"))
      ```
    """

    name: str
    """The name of the response format."""

    description: str | None
    """A description of the response format, if available."""

    schema: dict[str, object]
    """JSON schema representation of the structured output format."""

    mode: FormattingMode
    """The decorator-provided mode of the response format. 
    
    Determines how the LLM call may be modified in order to extract the expected format.
    """

    formatting_instructions: str | None
    """The formatting instructions that will be added to the LLM system prompt.

    If the format type has a `formatting_instructions` class method, the output of that
    call will be used for instructions. Otherwise, instructions may be auto-generated
    based on the formatting mode.
    """

    formattable: type[FormattableT]
    """The `Formattable` type that this `Format` describes.
    
    While the `FormattbleT` typevar allows for `None`, a `Format` will never be
    constructed when the `FormattableT` is `None`, so you may treat this as 
    a `RequiredFormattableT` in practice.
    """


@runtime_checkable
class HasFormattingInstructions(Protocol):
    """Protocol for classes that have been decorated with `@format()`."""

    @classmethod
    def formatting_instructions(cls) -> str | None: ...



"""The `Image` content class."""

import base64
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, get_args

import httpx

ImageMimeType = Literal[
    "image/png",
    "image/jpeg",
    "image/webp",
    "image/gif",
    "image/heic",
    "image/heif",
]  # TODO: add e2e tests for every supported type

MIME_TYPES = get_args(ImageMimeType)

# Maximum image size in bytes (20MB)
MAX_IMAGE_SIZE = 20 * 1024 * 1024


def infer_image_type(image_data: bytes) -> ImageMimeType:
    """Get the MIME type of an image from its raw bytes.

    Raises:
        ValueError: If the image type cannot be determined or data is too small
    """
    if len(image_data) < 12:
        raise ValueError("Image data too small to determine type (minimum 12 bytes)")

    if image_data.startswith(b"\xff\xd8\xff"):
        return "image/jpeg"
    elif image_data.startswith(b"\x89PNG\r\n\x1a\n"):
        return "image/png"
    elif image_data.startswith(b"GIF87a") or image_data.startswith(b"GIF89a"):
        return "image/gif"
    elif image_data.startswith(b"RIFF") and image_data[8:12] == b"WEBP":
        return "image/webp"
    elif image_data[4:12] in (
        b"ftypmif1",
        b"ftypmsf1",
        b"ftypheic",
        b"ftypheix",
        b"ftyphevc",
        b"ftyphevx",
    ):
        subtype = image_data[8:12]
        if subtype in (b"heic", b"heix"):
            return "image/heic"
        elif subtype in (b"mif1", b"msf1", b"hevc", b"hevx"):
            return "image/heif"
    raise ValueError("Unsupported image type")


@dataclass(kw_only=True)
class Base64ImageSource:
    """Image data represented as a base64 encoded string."""

    type: Literal["base64_image_source"]

    data: str
    """The image data, as a base64 encoded string."""

    mime_type: ImageMimeType
    """The mime type of the image (e.g. image/png)."""


def _process_image_bytes(data: bytes, max_size: int) -> Base64ImageSource:
    """Validate and process image bytes into a Base64ImageSource.

    Args:
        data: Raw image bytes
        max_size: Maximum allowed size in bytes

    Returns:
        A Base64ImageSource with validated and encoded data

    Raises:
        ValueError: If data size exceeds max_size
    """
    size = len(data)
    if size > max_size:
        raise ValueError(
            f"Image size ({size} bytes) exceeds maximum allowed size ({max_size} bytes)"
        )

    mime_type = infer_image_type(data)
    encoded_data = base64.b64encode(data).decode("utf-8")
    return Base64ImageSource(
        type="base64_image_source",
        data=encoded_data,
        mime_type=mime_type,
    )


@dataclass(kw_only=True)
class URLImageSource:
    """Image data referenced via external URL."""

    type: Literal["url_image_source"]

    url: str
    """The url of the image (e.g. https://example.com/sazed.png)."""


@dataclass(kw_only=True)
class Image:
    """Image content for a message.

    Images can be included in messages to provide visual context. This can be
    used for both input (e.g., user uploading an image) and output (e.g., model
    generating an image).
    """

    type: Literal["image"] = "image"

    source: Base64ImageSource | URLImageSource

    @classmethod
    def from_url(cls, url: str) -> "Image":
        """Create an `Image` reference from a URL, without downloading it.

        Args:
            url: The URL of the image

        Returns:
            An `Image` with a `URLImageSource`
        """
        return cls(source=URLImageSource(type="url_image_source", url=url))

    @classmethod
    def download(cls, url: str, *, max_size: int = MAX_IMAGE_SIZE) -> "Image":
        """Download and encode an image from a URL.

        Args:
            url: The URL of the image to download
            max_size: Maximum allowed image size in bytes (default: 20MB)

        Returns:
            An `Image` with a `Base64ImageSource`

        Raises:
            ValueError: If the downloaded image exceeds max_size
        """
        response = httpx.get(url, follow_redirects=True)
        response.raise_for_status()
        return cls(source=_process_image_bytes(response.content, max_size))

    @classmethod
    async def download_async(
        cls, url: str, *, max_size: int = MAX_IMAGE_SIZE
    ) -> "Image":
        """Asynchronously download and encode an image from a URL.

        Args:
            url: The URL of the image to download
            max_size: Maximum allowed image size in bytes (default: 20MB)

        Returns:
            An `Image` with a `Base64ImageSource`

        Raises:
            ValueError: If the downloaded image exceeds max_size
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(url, follow_redirects=True)
            response.raise_for_status()
            return cls(source=_process_image_bytes(response.content, max_size))

    @classmethod
    def from_file(cls, file_path: str, *, max_size: int = MAX_IMAGE_SIZE) -> "Image":
        """Create an `Image` from a file path.

        Args:
            file_path: Path to the image file
            max_size: Maximum allowed image size in bytes (default: 20MB)

        Raises:
            FileNotFoundError: If the file does not exist
            ValueError: If the file size exceeds max_size
        """
        path = Path(file_path)
        file_size = path.stat().st_size
        if file_size > max_size:
            raise ValueError(
                f"Image file size ({file_size} bytes) exceeds maximum allowed size ({max_size} bytes)"
            )
        with open(path, "rb") as f:
            image_bytes = f.read()
        return cls(source=_process_image_bytes(image_bytes, max_size))

    @classmethod
    def from_bytes(cls, data: bytes, *, max_size: int = MAX_IMAGE_SIZE) -> "Image":
        """Create an `Image` from raw bytes.

        Args:
            data: Raw image bytes
            max_size: Maximum allowed image size in bytes (default: 20MB)

        Raises:
            ValueError: If the data size exceeds max_size
        """
        return cls(source=_process_image_bytes(data, max_size))



"""Utilities for handling optional provider dependencies."""

from collections.abc import Callable


def create_import_error_stub(
    package_name: str, client_name: str
) -> Callable:  # pragma: no cover
    """Create a stub that raises ImportError when called.

    Args:
        package_name: The package/extra name (e.g., "anthropic", "openai", "google")
        client_name: The client name for the error message (e.g., "AnthropicClient")

    Returns:
        A callable that raises `ImportError` with helpful message.
    """

    def _raise_not_installed() -> None:
        raise ImportError(
            f"The '{package_name}' package is required to use {client_name}. "
            f"Install it with: `uv add 'mirascope[{package_name}]'`. "
            "Or use `uv add 'mirascope[all]'` to support all providers."
        )

    return _raise_not_installed


def create_client_stub(package_name: str, client_name: str) -> type:  # pragma: no cover
    """Create a stub client class that raises ImportError when instantiated.

    Args:
        package_name: The package/extra name (e.g., "anthropic", "openai", "google")
        client_name: The client name for the error message (e.g., "AnthropicClient")

    Returns:
        A stub class that raises `ImportError` on instantiation.
    """
    error_fn = create_import_error_stub(package_name, client_name)

    class _ClientStub:
        """Stub client that raises `ImportError` when instantiated."""

        def __init__(self) -> None:
            error_fn()

    return _ClientStub



"""Base parameters for LLM providers."""

from typing import TypedDict


class Params(TypedDict, total=False):
    """Common parameters shared across LLM providers.

    Note: Each provider may handle these parameters differently or not support them at all.
    Please check provider-specific documentation for parameter support and behavior.
    """

    temperature: float
    """Controls randomness in the output (0.0 to 1.0).

    Lower temperatures are good for prompts that require a less open-ended or
    creative response, while higher temperatures can lead to more diverse or
    creative results.
    """

    max_tokens: int
    """Maximum number of tokens to generate."""

    top_p: float
    """Nucleus sampling parameter (0.0 to 1.0).
    
    Tokens are selected from the most to least probable until the sum of their 
    probabilities equals this value. Use a lower value for less random responses and a
    higher value for more random responses.
    """

    top_k: int
    """Limits token selection to the k most probable tokens (typically 1 to 100).

    For each token selection step, the ``top_k`` tokens with the
    highest probabilities are sampled. Then tokens are further filtered based
    on ``top_p`` with the final token selected using temperature sampling. Use
    a lower number for less random responses and a higher number for more
    random responses.
    """

    seed: int
    """Random seed for reproducibility.
    
    When ``seed`` is fixed to a specific number, the model makes a best
    effort to provide the same response for repeated requests.

    Not supported by all providers, and does not guarantee strict reproducibility.
    """

    stop_sequences: list[str]
    """Stop sequences to end generation.
    
    The model will stop generating text if one of these strings is encountered in the
    response.
    """

    thinking: bool
    """Configures whether the model should use thinking.
    
    Thinking is a process where the model spends additional tokens thinking about the
    prompt before generating a response. You may configure thinking either by passing
    a bool to enable or disable it.

    If `params.thinking` is `True`, then thinking and thought summaries will be enabled
    (if supported by the model/provider), with a default budget for thinking tokens.

    If `params.thinking` is `False`, then thinking will be wholly disabled, assuming
    the model allows this (some models, e.g. `google:gemini-2.5-pro`, do not allow
    disabling thinking).

    If `params.thinking` is unset (or `None`), then we will use provider-specific default
    behavior for the chosen model.
    """

    encode_thoughts_as_text: bool
    """Configures whether `Thought` content should be re-encoded as text for model consumption.
    
    If `True`, then when an `AssistantMessage` contains `Thoughts` and is being passed back
    to an LLM, those `Thoughts` will be encoded as `Text`, so that the assistant can read
    those thoughts. That ensures the assistant has access to (at least the summarized output of)
    its reasoning process, and contrasts with provider default behaviors which may ignore
    prior thoughts, particularly if tool calls are not involved.

    When `True`, we will always re-encode Mirascope messages being passed to the provider,
    rather than reusing raw provider response content. This may disable provider-specific
    behavior like cached reasoning tokens.

    If `False`, then `Thoughts` will not be encoded as text, and whether reasoning context
    is available to the model depends entirely on the provider's behavior. 

    Defaults to `False` if unset.
    """
